package ru.dbotthepony.mc.otm.core

import java.util.*
import kotlin.ConcurrentModificationException

class TickList {
	private val conditional = LinkedList<IConditionalTickable>()
	private val once = LinkedList<ITickable>()

	private val conditionalValveTime = LinkedList<IConditionalTickable>()
	private val onceValveTime = LinkedList<ITickable>()

	private var inTicker = false

	fun add(ticker: IConditionalTickable) {
		if (inTicker) {
			conditionalValveTime.add(ticker)
		} else {
			conditional.addFirst(ticker)
		}
	}

	fun add(ticker: ITickable) {
		if (inTicker) {
			onceValveTime.add(ticker)
		} else {
			once.addFirst(ticker)
		}
	}

	fun tick() {
		if (inTicker) {
			throw ConcurrentModificationException("Already ticking")
		}

		inTicker = true
		val iterator = conditional.iterator()

		for (ticker in iterator) {
			if (!ticker.canTick) {
				iterator.remove()
			} else {
				ticker.tick()
			}
		}

		for (ticker in once) {
			ticker.tick()
		}

		once.clear()

		for (ticker in conditionalValveTime) {
			conditional.addFirst(ticker)
		}

		for (ticker in onceValveTime) {
			once.addFirst(ticker)
		}

		conditionalValveTime.clear()
		onceValveTime.clear()

		inTicker = false
	}

	fun clear() {
		conditional.clear()
		once.clear()
	}
}

fun interface ITickable {
	fun tick()
}

interface IConditionalTickable : ITickable {
	/**
	 * Once this returns false, it should stay false.
	 *
	 * If it suddenly turns true after being false, result is undefined.
	 */
	val canTick: Boolean
}
